﻿using FrameworkLibrary.System.Windows.ComponentModel;
using FrameworkLibrary.System.Windows.Input;
using Microsoft.Win32;
using Prototype.Model;
using Prototype.Repros;
using Syncfusion.Windows.Controls.Grid.Converter;
using Syncfusion.Windows.Controls.Grid;
using Syncfusion.XlsIO;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Threading;
using Prototype.Helpers;
using System.IO;

namespace Prototype.ViewModels
{
    public class CrudViewModelBase<T1> : IsBusyViewModel, ITitleViewModel, IDisposable where T1 : class, IEntity
    {
        #region fields
        private UnitOfWork _UnitOfWork;
        private bool _Disposed = false;
        private string _Title;
        private string _RepositoryName;
        private IRepository<T1> _Repository;

        private ObservableCollection<T1> _Data;
        private T1 _SelectedItem;
        private Visibility _ExcelImportVisibility;
        private Visibility _NewVisibility;
        private Visibility _CloneVisibility;
        private Visibility _EditVisibility;
        private Visibility _DeleteVisibility;
        private Visibility _SaveVisibility;
        private Visibility _RefreshVisibility;
        private Visibility _ResetFilterVisibility;
        private Visibility _ExcelExportVisibility;
        private RelayCommand _ExcelImportCommand;
        private RelayCommand _NewCommand;
        private RelayCommand _CloneCommand;
        private RelayCommand _EditCommand;
        private RelayCommand _DeleteCommand;
        private RelayCommand _SaveCommand;
        private RelayCommand _RefreshCommand;
        private RelayCommand _ResetFilterCommand;
        private RelayCommand _ExcelExportCommand;
        #endregion

        #region constructors
        protected CrudViewModelBase(string title, string repositoryName)
        {
            if (string.IsNullOrEmpty(title)) throw new ArgumentNullException("title");
            if (string.IsNullOrEmpty(repositoryName)) throw new ArgumentNullException("repositoryName");
            this.Title = title;
            this.RepositoryName = repositoryName;
            this.UnitOfWork = new UnitOfWork(new EDSDBEntities());
            this.MySetRepository();
            this.Data = new ObservableCollection<T1>();
            this.ExcelImportCommand = new RelayCommand(ExecuteExcelImportCommand, CanExecuteExcelImportCommand);
            this.NewCommand = new RelayCommand(ExecuteNewCommand, CanExecuteNewCommand);
            this.CloneCommand = new RelayCommand(ExecuteCloneCommand, CanExecuteCloneCommand);
            this.EditCommand = new RelayCommand(ExecuteEditCommand, CanExecuteEditCommand);
            this.DeleteCommand = new RelayCommand(ExecuteDeleteCommand, CanExecuteDeleteCommand);
            this.SaveCommand = new RelayCommand(ExecuteSaveCommand, CanExecuteSaveCommand);
            this.RefreshCommand = new RelayCommand(ExecuteRefreshCommand, CanExecuteRefreshCommand);
            this.ResetFilterCommand = new RelayCommand(ExecuteResetFilterCommand, CanExecuteResetFilterCommand);
            this.ExcelExportCommand = new RelayCommand(ExecuteExcelExportCommand, CanExecuteExcelExportCommand);
        }
        #endregion


        #region properties

        public ObservableCollection<T1> Data
        {
            get { return _Data; }
            set { _Data = value; }
        }

        public T1 SelectedItem
        {
            get { return _SelectedItem; }
            set
            {
                _SelectedItem = value;
                CommandManager.InvalidateRequerySuggested();
                this.NotifyPropertyChanged("SelectedItem");
            }
        }



        public UnitOfWork UnitOfWork
        {
            get { return _UnitOfWork; }
            private set { _UnitOfWork = value; }
        }

        public bool Disposed
        {
            get { return _Disposed; }
            private set { _Disposed = value; }
        }

        public string Title
        {
            get { return _Title; }
            set
            {
                _Title = value;
                this.NotifyPropertyChanged("Title");
            }
        }

        public string RepositoryName
        {
            get { return _RepositoryName; }
            set { _RepositoryName = value; }
        }


        public IRepository<T1> Repository
        {
            get { return _Repository; }
            set
            { _Repository = value;
                if(_Repository != null)
                {
                    _Repository.AddedOrRemoved += (object sender, RepositoryAddedOrRemovedEventArgs args) =>
                    {
                        CommandManager.InvalidateRequerySuggested();
                        var dummy = this.GetHasChanges();
                    };
                }
            }
        }


        #endregion

        #region visibility properties
        public Visibility ExcelImportVisibility
        {
            get { return _ExcelImportVisibility; }
            set
            { _ExcelImportVisibility = value;
              this.NotifyPropertyChanged("ExcelImportVisibility");
            }
        }
        public Visibility NewVisibility
        {
            get { return _NewVisibility; }
            set
            { _NewVisibility = value;
              this.NotifyPropertyChanged("NewVisibility");
            }
        }

        public Visibility CloneVisibility
        {
            get { return _CloneVisibility; }
            set
            { _CloneVisibility = value;
              this.NotifyPropertyChanged("CloneVisibility");
            }
        }

        public Visibility EditVisibility
        {
            get { return _EditVisibility; }
            set
            { _EditVisibility = value;
              this.NotifyPropertyChanged("EditVisibility");
            }
        }

        public Visibility DeleteVisibility
        {
            get { return _DeleteVisibility; }
            set
            { _DeleteVisibility = value;
              this.NotifyPropertyChanged("DeleteVisibility");
            }
        }

        public Visibility SaveVisibility
        {
            get { return _SaveVisibility; }
            set
            { _SaveVisibility = value;
              this.NotifyPropertyChanged("SaveVisibility");
            }
        }

        public Visibility RefreshVisibility
        {
            get { return _RefreshVisibility; }
            set
            { _RefreshVisibility = value;
              this.NotifyPropertyChanged("RefreshVisibility");
            }
        }

        public Visibility ResetFilterVisibility
        {
            get { return _ResetFilterVisibility; }
            set
            { _ResetFilterVisibility = value;
              this.NotifyPropertyChanged("ResetFilterVisibility");
            }
        }

        public Visibility ExcelExportVisibility
        {
            get { return _ExcelExportVisibility; }
            set
            { _ExcelExportVisibility = value;
              this.NotifyPropertyChanged("ExcelExportVisibility");
            }
        }
        #endregion
        #region commands
        public RelayCommand ExcelImportCommand
        {
            get { return _ExcelImportCommand; }
            set
            {
                _ExcelImportCommand = value;
            }
        }
        public RelayCommand NewCommand
        {
            get { return _NewCommand; }
            set
            {
                _NewCommand = value;
            }
        }
        public RelayCommand CloneCommand
        {
            get { return _CloneCommand; }
            set
            {
                _CloneCommand = value;
            }
        }
        public RelayCommand EditCommand
        {
            get { return _EditCommand; }
            set
            {
                _EditCommand = value;
            }
        }
        public RelayCommand DeleteCommand
        {
            get { return _DeleteCommand; }
            set
            {
                _DeleteCommand = value;
            }
        }
        public RelayCommand SaveCommand
        {
            get { return _SaveCommand; }
            set
            {
                _SaveCommand = value;
            }
        }
        public RelayCommand RefreshCommand
        {
            get { return _RefreshCommand; }
            set
            {
                _RefreshCommand = value;
            }
        }
        public RelayCommand ResetFilterCommand
        {
            get { return _ResetFilterCommand; }
            set
            {
                _ResetFilterCommand = value;
            }
        }
        public RelayCommand ExcelExportCommand
        {
            get { return _ExcelExportCommand; }
            set
            {
                _ExcelExportCommand = value;
            }
        }
        #endregion

        #region methods
        public bool GetHasChanges()
        {
            return this.MyGetHasChanges();
        }
        #endregion


        #region overridable command methods
        public virtual bool OnCheckHasDisplayRights(UserControl userControl)
        {
            if (userControl == null) throw new ArgumentNullException("userControl");
            bool hasDisplayRights = true;
            return hasDisplayRights;
        }
        public virtual void OnLoadData()
        {

        }
        public virtual void OnExcelImport(Object param)
        {
            MessageBox.Show("ExcelImport");
        }
        public virtual void OnNew(Object param)
        {
            if(!this.Disposed)
            {
                Type entityType = typeof(T1);
                T1 entity = default(T1);
                entity = (T1)Activator.CreateInstance(entityType);
                entity.IsNew = true;
                entity.SetDefaultValues();
                this.Repository.Add(entity);
                this.Data.Add(entity);

                this.SelectedItem = entity;
                CommandManager.InvalidateRequerySuggested();
            }
            //MessageBox.Show("New");
        }
        public virtual void OnClone(Object param)
        {
            if((!this.Disposed) &&(this.SelectedItem != null))
            {
                T1 entity = (T1)this.SelectedItem.Clone();
                entity.IsNew = true;
                entity.UID = Guid.NewGuid();
                entity.CreatedAt = null;
                entity.CreatedFrom = null;
                entity.LastModifiedAt = null;
                entity.LastModifiedFrom = null;
                this.Repository.Add(entity);
                this.Data.Add(entity);

                this.SelectedItem = entity;
            }

            CommandManager.InvalidateRequerySuggested();
            //MessageBox.Show("Clone");
        }
        public virtual void OnEdit(Object param)
        {
            MessageBox.Show("Edit");
        }
        public virtual void OnDelete(Object param)
        {
            if ((!this.Disposed) && (this.SelectedItem != null))
            {
                T1 delEntity = (T1)this.SelectedItem;

                this.Repository.Remove(delEntity);
                int index = this.Data.IndexOf(delEntity);
                this.Data.Remove(delEntity);
                T1 selEntity = null;
                index++;
                if (index < this.Data.Count())
                {
                    selEntity = this.Data[index];
                }
                else
                {
                    index--;
                    index--;
                    if(index > 0)
                    {
                        selEntity = this.Data[index];
                    }
                }
                if(selEntity != null)
                {
                    this.SelectedItem = selEntity;
                }

            }
            CommandManager.InvalidateRequerySuggested();
            //MessageBox.Show("Delete");
        }
        public virtual async Task<int> OnSaveAsync(Object param)
        {
            int returnValue = 0;

            if(!this.Disposed)
            {
                Exception exception = null;
                bool failed = false;
                this.IsBusyText = "speichere Änderungen";
                this.IsBusy = true;
                try
                {
                    returnValue = await this.UnitOfWork.CompleteAsync();
                }
                catch (Exception ex)
                {
                    exception = ex;
                    failed = (exception != null);
                }
                this.IsBusyText = string.Empty;
                this.IsBusy = false;
                if(failed)
                {
                    MessageBox.Show("Beim Speichern der Änderungen trat ein Problem auf:" + exception.Message);
                }
                else
                {
                    MessageBox.Show("Änderungen gespeichert.");
                }
            }
            return returnValue;
        }
        public virtual void OnSave(Object param)
        {

            if (!this.Disposed)
            {
                Exception exception = null;
                bool failed = false;
                this.IsBusyText = "speichere Änderungen";
                this.IsBusy = true;
                try
                {
                    this.UnitOfWork.Complete();
                }
                catch (Exception ex)
                {
                    exception = ex;
                    failed = (exception != null);
                }
                this.IsBusyText = string.Empty;
                this.IsBusy = false;
                if (failed)
                {
                    MessageBox.Show("Beim Speichern der Änderungen trat ein Problem auf:" + exception.Message);
                }
                else
                {
                    MessageBox.Show("Änderungen gespeichert.");
                }
            }
        }
        public virtual  void OnRefresh(Object param)
        {
            if ((!this.Disposed) && (this.GetHasChanges()))
            {
                MessageBoxResult messageBoxResult = MessageBox.Show("Es existieren ungespeicherte Änderungen. Sollen diese vor der Aktualisierung gespeichert werden?", "ungespeicherte Änderungen", MessageBoxButton.YesNoCancel);
                if (messageBoxResult == MessageBoxResult.Yes)
                {
                     this.OnSave(null);
                }
            }

            //MessageBox.Show("Refresh");
            this.OnLoadData();
        }
        public virtual void OnResetFilter(Object param)
        {
            if((param != null) && (param is GridDataControl))
            {
                GridDataControl gridDataControl = (GridDataControl)param;
                //this.dataGrid1.VisibleColumns[“CustomerID”].Filters.Clear();
                foreach (GridDataVisibleColumn loopGridDataVisibleColumn in gridDataControl.VisibleColumns)
                {
                    loopGridDataVisibleColumn.Filters.Clear();
                }

            }

            //MessageBox.Show("ResetFilter");
        }
        public virtual void OnExcelExport(Object param)
        {
            if ((!this.Disposed) &&(param != null) && (param is GridDataControl))
            {
                GridDataControl gridDataControl = (GridDataControl)param;
                SaveFileDialog saveFileDialog = new SaveFileDialog();
                saveFileDialog.Filter = "Excel |*.xlsx";
                saveFileDialog.Title = "Excel Export speichern unter";
                saveFileDialog.FileName = (this.Title + ".xlsx");
                saveFileDialog.ShowDialog();
                if (!string.IsNullOrEmpty(saveFileDialog.FileName))
                {
                    string fileName = saveFileDialog.FileName;
                    bool success = false;
                    try
                    {
                        //gridDataControl.ExportToExcel(fileName, ExcelVersion.Excel97to2003, exportingHandler);
                        gridDataControl.ExportToExcel(fileName, ExcelVersion.Excel2016, exportingHandler);
                        //TODO BK FROM BK  check fileexistence here.
                        if (!File.Exists(fileName))
                        {
                            throw new FileNotFoundException(string.Format("die Datei '{0}' wurde nicht erstellt.", fileName));
                        }
                        MessageBox.Show(string.Format("die Daten wurden in die Datei '{0}' exportiert.", fileName));
                        success = true;
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(string.Format("Beim Exportieren der Daten in die Datei '{0}' ist ein Fehler aufgetreten:", fileName) + ex.Message.ToString());
                        MessageBox.Show(ex.Message.ToString());
                    }
                    if(success)
                    {
                        MessageBoxResult messageBoxResult = MessageBox.Show("Soll die Exceldatei geöffnet werden?", "Datei öffnen", MessageBoxButton.YesNoCancel);
                        if(messageBoxResult == MessageBoxResult.Yes)
                        {
                            System.Diagnostics.Process.Start(fileName);
                        }
                    }
                }
            }
            //MessageBox.Show("ExcelExport");
        }
        private void exportingHandler(object sender, ExportingToExcelEventArgs e)
        {
            if (e.Style.CellType == "ComboBox")
            {
                GridStyleInfo GridStyleInfo = e.Style;

                var cellValue = GridStyleInfo.CellValue;
                string text = GridStyleInfo.Text;
            }
        }

        public virtual void OnDiscardChanges(Object param)
        {
            this.UnitOfWork.DiscardChanges();
            CommandManager.InvalidateRequerySuggested();
            this.OnRefresh(param);
        }
        #endregion

        #region command methods
        private bool CanExecuteExcelImportCommand(Object param)
        {
            return true;
        }
        private void ExecuteExcelImportCommand(Object param)
        {
            this.OnExcelImport(param);
        }
        private bool CanExecuteNewCommand(Object param)
        {
            return true;
        }
        private void ExecuteNewCommand(Object param)
        {
            this.OnNew(param);
        }
        private bool CanExecuteCloneCommand(Object param)
        {
            return (this.SelectedItem != null);
        }
        private void ExecuteCloneCommand(Object param)
        {
            this.OnClone(param);
        }
        private bool CanExecuteEditCommand(Object param)
        {
            return true;
        }
        private void ExecuteEditCommand(Object param)
        {
            this.OnEdit(param);
        }
        private bool CanExecuteDeleteCommand(Object param)
        {
            return (this.SelectedItem != null);
        }
        private void ExecuteDeleteCommand(Object param)
        {
            this.OnDelete(param);
        }
        private bool CanExecuteSaveCommand(Object param)
        {
            return this.MyGetHasChanges();
        }
        private async void ExecuteSaveCommand(Object param)
        {
            await this.OnSaveAsync(param);
        }
        private bool CanExecuteRefreshCommand(Object param)
        {
            return true;
        }
        private void ExecuteRefreshCommand(Object param)
        {
            this.OnRefresh(param);
        }
        private bool CanExecuteResetFilterCommand(Object param)
        {
            return true;
        }
        private void ExecuteResetFilterCommand(Object param)
        {
            this.OnResetFilter(param);
        }
        private bool CanExecuteExcelExportCommand(Object param)
        {
            return true;
        }
        private void ExecuteExcelExportCommand(Object param)
        {
            this.OnExcelExport(param);
        }
        #endregion
        #region IDisposable methods
        public void Dispose()
        {
            MyDispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion

        #region private methods
        private void MyDispose(bool disposing)
        {
            if (this.Disposed)
                return;

            if (disposing)
            {
                this.UnitOfWork.Dispose();
            }

            this.Disposed = true;
        }
        private bool MyGetHasChanges()
        {
            return ((!this.Disposed) &&(this.UnitOfWork.GetHasChanges()));

        }
        private void MySetRepository()
        {
            IRepository<T1> repository = null;           
            Type unitOfWorkType = this.UnitOfWork.GetType();
            PropertyInfo propertyInfo = unitOfWorkType.GetProperty(this.RepositoryName);
            if (propertyInfo != null)
            {
                repository = (IRepository < T1 >)propertyInfo.GetValue(this.UnitOfWork);
            }

            if (repository == null)
            {
                throw new InvalidOperationException(string.Format("es konnte kein passendes Repository '{0}' gefunden werden.", this.RepositoryName));
            }
            else
            {
                this.Repository = repository;
            }

        }

        #endregion
    }
}
